
import numpy as np, scipy as sp, itertools as it, collections, math, time, enum
import scipy.linalg

import McUtils.Numputils as nput
from McUtils.Scaffolding import Logger
from McUtils.Formatters import TableFormatter
from McUtils.Zachary import DensePolynomial
from McUtils.Combinatorics import IntegerPartitioner, IntegerPartitioner2D, UniquePermutations

from ..BasisReps import StateMaker
from ..Modes import NormalModes

__all__ = [
    'FranckCondonModel'
]

class State(enum.Enum):
    GroundState = 'gs'
    ExcitedState = 'es'

    @classmethod
    def get_aliases(cls):
        return {
            'ground-state': 'gs',
            'excited-state': 'es'
        }

    @classmethod
    def resolve(cls, method_name):
        if isinstance(method_name, State):
            return method_name
        method_name = method_name.lower()
        method_name = cls.get_aliases().get(method_name, method_name)
        return cls(method_name)

class RotationMethods(enum.Enum):
    LeastSquares = 'least-squares'
    Duschinsky = 'duschinsky'

    @classmethod
    def get_aliases(cls):
        return {
            'default': 'duschinsky',
            'lstsq': 'least-squares'
        }
    @classmethod
    def resolve(cls, method_name):
        if isinstance(method_name, RotationMethods):
            return method_name
        method_name = method_name.lower()
        method_name = cls.get_aliases().get(method_name, method_name)
        return cls(method_name)


class FranckCondonModel:

    # @classmethod
    # def get_simple_poly_evaluator(cls, exponents):
    #     """
    #     Provides a simple function for evaluating based on the exponents without rotation
    #     (i.e. product of 1D integrals)
    #
    #     :param exponents:
    #     :return:
    #     """
    #

    default_rotation_method = 'default'
    default_rotation_order = State.GroundState
    default_rotation_center = State.ExcitedState
    default_embedding_ref = State.ExcitedState
    default_include_rotation = True
    def __init__(self, gs_nms: NormalModes, es_nms, atoms=None, *,
                 logger=None,
                 embed=True, embedding_ref=None,
                 masses=None,
                 mass_weight=True,
                 dimensionless=False,
                 mode_selection=None,
                 mode_reordering=None,
                 rotation_method=None,
                 rotation_order=None,
                 rotation_center=None,
                 include_rotation=None,
                 rotation_blocks=None
                 ):

        self.gs_nms, self.es_nms, _ = self.prep_modes(
            gs_nms, es_nms,
            embed=embed, embedding_ref=embedding_ref,
            masses=masses,
            mass_weight=mass_weight,
            dimensionless=dimensionless,
            mode_selection=mode_selection,
            mode_reordering=mode_reordering,
        )

        self.rotation_method = rotation_method
        self.rotation_order = rotation_order
        self.rotation_center = rotation_center
        self.include_rotation = include_rotation
        self.rotation_blocks = rotation_blocks
        self.atoms = atoms

        self.logger = Logger.lookup(logger)

    @classmethod
    def prep_modes(cls, gs_nms, es_nms,
                   embed=True, embedding_ref=None,
                   masses=None,
                   mass_weight=False,
                   dimensionless=False,
                   mode_selection=None,
                   mode_reordering=None,
                   **rotation_opts
                   ):
        if embed:
            if embedding_ref is None:
                embedding_ref = cls.default_embedding_ref
            gs_nms, es_nms = cls.embed_modes(gs_nms, es_nms, masses=masses, ref=embedding_ref)

        if mass_weight:
            gs_nms = cls.mass_weight_nms(gs_nms, masses=masses)
            es_nms = cls.mass_weight_nms(es_nms, masses=masses)

        if dimensionless:
            gs_nms = cls.make_dimensionless(gs_nms, masses=masses)
            es_nms = cls.make_dimensionless(es_nms, masses=masses)

        if mode_selection is not None:
            gs_nms = gs_nms[mode_selection]
            es_nms = es_nms[mode_selection]

        if mode_reordering is not None:
            # gs_nms = gs_nms[mode_selection]
            es_nms = es_nms[mode_reordering]

        return gs_nms, es_nms, rotation_opts

    @classmethod
    def from_files(cls, gs_file, es_file,
                   logger=None, mode_selection=None,
                   mode_reordering=None,
                   internals=None, internals_ref='gs',
                   **rotation_embedding_opts):
        from ..Molecools import Molecule
        internals_ref = State.resolve(internals_ref)
        if internals_ref == State.GroundState:
            gs_mol = Molecule.from_file(gs_file, internals=internals)
            ics = gs_mol.internal_coordinates  # set up the reference
            es_mol = Molecule.from_file(es_file, internals=gs_mol.internals)
        else:
            es_mol = Molecule.from_file(es_file, internals=internals)
            ics = es_mol.internal_coordinates  # set up the reference
            gs_mol = Molecule.from_file(gs_file, internals=es_mol.internals)
        return cls.from_mols(
            gs_mol,
            es_mol,
            logger=logger,
            mode_selection=mode_selection,
            mode_reordering=mode_reordering,
            **rotation_embedding_opts
        )

    @classmethod
    def convert_internal_modes(cls, mol, nms):
        raise NotImplementedError("ignored")
        RX = mol.get_cartesians_by_internals(1, strip_embedding=True, reembed=True)[0]
        XR = mol.get_internals_by_cartesians(1, strip_embedding=True)[0]

        mat = RX @ nms.modes_by_coords # RX @ XQ
        inv = nms.coords_by_modes @ XR # QX @ XR

        # print(inv @ mat) #, nms.matrix @ nms.inverse)
        # raise Exception(...)
        freqs = nms.freqs

        return NormalModes(
            mol.internal_coordinates.system,
            mat,
            inverse=inv,
            freqs=freqs,
            origin=mol.internal_coordinates
        )

    @classmethod
    def from_mols(cls, gs, es,
                  logger=None,
                  remove_transrot=True,
                  use_internals=True,
                  embed=True,
                  mass_weight=True,
                  **rotation_embedding_opts):
        use_internals = use_internals and (gs.internals is not None)
        if use_internals:
            gg = NormalModes.from_molecule(gs, use_internals=True)
            ee = NormalModes.from_molecule(es, use_internals=True)
        elif remove_transrot:
            gg = NormalModes.from_molecule(gs, project_transrot=True, mass_weighted=True, use_internals=False)
            ee = NormalModes.from_molecule(es, project_transrot=True, mass_weighted=True, use_internals=False)
        else:
            gg = gs.normal_modes.modes.basis.to_new_modes()
            ee = es.normal_modes.modes.basis.to_new_modes()
        # if use_internals:
        #     # es = gs.get_embedded_molecule(ref=gs)
        #     # gg, ee = cls.embed_modes(gg, ee, masses=gs.atomic_masses, ref='gs')
        #     gg = cls.convert_internal_modes(gs, gg)
        #     ee = cls.convert_internal_modes(es, ee)
        #
        #     raise Exception(gg.inverse @ gg.matrix)

        return cls(
            gg, ee,
            atoms=gs.atoms,
            embed=not use_internals and embed,
            logger=logger,
            **rotation_embedding_opts
        )

    def get_overlaps(self, excitations, *, duschinsky_cutoff=None, ground_states=None,
                     return_states=True,
                     **rotation_opts
                     ):
        gs_nms, es_nms, rotation_opts = self.prep_modes(
            self.gs_nms, self.es_nms,
            **rotation_opts
        )

        excitations = self.prep_state_space(excitations, es_nms)
        ground_states = self.prep_state_space(ground_states, gs_nms)

        ret = self.get_fcfs(
            gs_nms, es_nms,
            excitations,
            **self.prep_opts(
                ground_states=ground_states,
                duschinsky_cutoff=duschinsky_cutoff,
                **rotation_opts
            )
        )

        if return_states:
            ret = (ret, (ground_states, excitations))

        return ret


    OverlapData = collections.namedtuple('OverlapData', ['alphas', 'scaling', 'center', 'gs', 'es'])
    Embedding = collections.namedtuple('Embedding', ['modes', 'center'])
    def get_overlap_data(self, **rotation_embedding_opts) -> OverlapData:
        opts = self.prep_opts(**rotation_embedding_opts)
        gs_nms, es_nms, rotation_opts = self.prep_modes(self.gs_nms, self.es_nms, **opts)
        gs_data, es_data = self.prep_overlap_args(self.gs_nms, self.es_nms)

        (alphas, modes_c, center, scaling), (L_gs, c_gs), (L_es, c_es) = self.get_overlap_gaussian_data(
            *gs_data,
            *es_data,
            **rotation_opts
        )

        return self.OverlapData(
            alphas,
            scaling,
            self.Embedding(modes_c, center),
            self.Embedding(L_gs, c_gs),
            self.Embedding(L_es, c_es)
        )

    _rot_opts = {'rotation_center', 'rotation_order', 'include_rotation', 'rotation_method'}
    @classmethod
    def prep_overlap_args(self, gs_nms, es_nms):
        # gs_nms = gs_nms.make_mass_weighted()
        # es_nms = es_nms.make_mass_weighted()

        freqs_gs = np.asanyarray(gs_nms.freqs)
        if gs_nms.frequency_scaled:
            freqs_gs = np.ones_like(freqs_gs)
        freqs_es = np.asanyarray(es_nms.freqs)
        if es_nms.frequency_scaled:
            freqs_es = np.ones_like(freqs_es)
        modes_gs = np.asanyarray(gs_nms.modes_by_coords)
        modes_es = np.asanyarray(es_nms.modes_by_coords)
        inv_gs = np.asanyarray(gs_nms.coords_by_modes)
        inv_es = np.asanyarray(es_nms.coords_by_modes)
        center_gs = np.asanyarray(gs_nms.origin).flatten()
        center_es = np.asanyarray(es_nms.origin).flatten()

        return (
            (freqs_gs, modes_gs, inv_gs, center_gs),
            (freqs_es, modes_es, inv_es, center_es),
        )

    @classmethod
    def get_poly_evaluation_plan(self, exponents, alphas=None, zpe_prod=None):
        """
        Provides a function that can take a set of indices and rotation matrices
        from the gs and es bases to the shared basis of the central Gaussian and compute
        the corresponding term contributions by considering every even
        permutation of indices that could lead to a non-zero contribution

        :param tg:
        :param te:
        :return:
        """
        key = exponents
        exponents = np.asanyarray(exponents)
        n = np.sum(exponents, dtype=int)

        max_gamma = self.df_weights(n) # we only need to compute this once

        evaluators = []
        if n == 0:
            def evaluate_zero(alphas, zpe_prod, exped_poly_coeffs, batch_size=int(1e7)):
                base_val = self.zero_point_alpha_contrib(alphas) if zpe_prod is None else zpe_prod
                if not (isinstance(exped_poly_coeffs[0], np.ndarray) and exped_poly_coeffs[0].ndim == 2):
                    base_val = np.full(len(exped_poly_coeffs[0]), base_val)
                return base_val
            evaluators.append(evaluate_zero)
        else:

            for parts in IntegerPartitioner.partitions(n): # arrays of partitions with the same total order
                good_parts = np.all(parts % 2 == 0, axis=1) # only even order partitions make sense
                parts = parts[good_parts]
                if len(parts) == 0: continue

                # find the different numbers of ways to split the terms up
                term_classes = [
                    IntegerPartitioner2D.get_partitions(exponents, terms)#[:1]
                    for terms in parts
                ]

                term_class_inds = [
                    [
                        (slice(None),) + np.where(np.ones(t.shape)) + (t.flatten(),)
                        for t in term_class
                    ]
                    for term_class in term_classes
                ]

                weights = [
                    np.array([
                        np.prod([UniquePermutations.count_permutations(counts) for counts in term])
                        for term in term_class
                        ])
                    for term_class in term_classes
                ]

                evaluators.append(
                    self.term_evaluator(parts, term_classes, term_class_inds, weights, max_gamma, alphas=alphas, zpe_prod=zpe_prod)
                )

        def evaluate_contrib(alphas, zpe_prod, poly_coeffs, batch_size=int(1e7)):
            return sum(
                [
                    ev(alphas, zpe_prod, poly_coeffs, batch_size=batch_size)
                    for ev in evaluators
                 ]
            )

        return evaluate_contrib

    @classmethod
    def zero_point_alpha_contrib(cls, alphas):
        return (np.sqrt(2 * np.pi) ** len(alphas)) / np.prod(np.sqrt(alphas))
    @classmethod
    def term_evaluator(self, exponents_list, splits_list, splits_inds_list, weights_list, gammas, alphas=None, zpe_prod=None):
        ndim = len(exponents_list[0])
        prefacs = np.array([np.prod(gammas[exponents]) for exponents in exponents_list])
        if alphas is not None:
            prefacs = prefacs * (self.zero_point_alpha_contrib(alphas) if zpe_prod is None else zpe_prod)
        prealphas = alphas

        def evaluate_contrib(alphas, zpe_prod, exped_poly_coeffs, batch_size=int(1e7)):
            contrib = [0] * len(exponents_list)

            scaling = prefacs * (
                1
                    if prealphas is not None else
                self.zero_point_alpha_contrib(alphas)
                    if zpe_prod is None else
                zpe_prod
            )

            ncoords = len(alphas)
            nperms = math.comb(ncoords, ndim)
            # try to split evenly into k batches
            num_batches = 1 + (nperms // batch_size)
            real_batch_size = nperms // num_batches
            batch_remainder = nperms % real_batch_size
            # # see how well we can split the remainder over the batches
            # subremainder = batch_remainder % num_batches
            # real_batch_size += batch_remainder // num_batches
            # batch_remainder = subremainder

            # print(":::::", exponents_list, ":::::")
            # print("    scaling:", scaling)

            batch_generator = it.combinations(range(ncoords), ndim)
            for batch_number in range(num_batches):
                num_terms = real_batch_size
                if batch_remainder > 0: # exhaust remainder bit by bit
                    num_terms += 1
                    batch_remainder -= 1
                ind_array = np.empty((num_terms, ndim), dtype=int)
                for k,ind in enumerate(it.islice(batch_generator, num_terms)):
                    ind_array[k] = ind

                new_contribs = self.evaluate_poly_contrib_chunk(
                    ind_array,
                    exponents_list,
                    splits_list,
                    splits_inds_list,
                    weights_list,
                    alphas,
                    exped_poly_coeffs
                )
                for i,c in enumerate(new_contribs):
                    contrib[i] += c#sum(cc*ss for cc,ss in zip(c, scaling))
                    # if isinstance(c, list):
                    #     contrib[i] += sum(c)
                    #     # if not isinstance(contrib[i], list):
                    #     #     if contrib[i] != 0: raise ValueError("mismatch in contribs...")
                    #     #     contrib[i] = c
                    #     # else:
                    #     #     for j,cc in enumerate(c):
                    #     #         contrib[i][j] += cc
                    # else:
                    #     contrib[i] += c

            cont = sum(c * s for c,s in zip(contrib, scaling))
            # print("     contrib:", cont)
            return cont

        return evaluate_contrib

    @staticmethod
    def _expand(pc, splits):
        return pc[:, np.newaxis, :, :] ** splits[np.newaxis, :, :, :]  # c x l x k x d
    @staticmethod
    def _contract(poly_exps, weights):
        return np.dot(
                np.prod(np.prod(poly_exps, axis=-1), axis=-1),
                weights
            )
    @classmethod
    def _index_contract(cls, pc, weights, split_inds):
        return sum(
            w * np.prod(pc[s], axis=-1)
            for w, s in zip(weights, split_inds)
        )
    @classmethod
    def evaluate_poly_chunks(cls, poly_coeffs, exps, splits, split_inds, weights, alphas, include_baseline=False):
        if isinstance(poly_coeffs, np.ndarray) and poly_coeffs.ndim == 2:
            return cls.evaluate_poly_chunks([poly_coeffs], exps, splits, split_inds, weights, alphas)[0]

        if include_baseline:
            raise NotImplementedError("need to add")

        alpha_contrib = np.prod(1 / (alphas ** (exps[np.newaxis] // 2)), axis=-1)  # c array of values
        # compute contribution from polynomial factors, can add them up immediately
        multi_contrib = np.zeros(len(poly_coeffs))
        for ii,pc in enumerate(poly_coeffs):
            # poly_exps = cls._expand(pc[:, :, :, 1], splits)
            # poly_contrib = cls._contract(poly_exps, weights)
            poly_contrib = cls._index_contract(pc, weights, split_inds)
            multi_contrib[ii] = np.dot(alpha_contrib, poly_contrib)

        return multi_contrib

    @classmethod
    def evaluate_poly_contrib_chunk(self, inds, exponents_list, splits_list, splits_inds_list,
                                    weights_list, alphas, coeffs):
        # k is the number of initial coords, d is the number of final coords
        # n is the total number of coords
        # exps is a d vector of exponents
        # splits is an l x k x d array of the way the exps divide over the initial coords
        # coeffs is an k x n x e array of exponentiated polynomial coeffs for each coord
        # inds is a c x d array where c is the number of inds in this chunk
        # alphas is a n vector

        multicoeff = not isinstance(coeffs, np.ndarray) and isinstance(coeffs[0], np.ndarray)

        e = coeffs.shape[-1] if not multicoeff else coeffs[0].shape[-1]
        c = inds.shape[0]
        k = coeffs.shape[1] if not multicoeff else coeffs[0].shape[1]
        d = inds.shape[1]
        if multicoeff:
            poly_coeffs = []
            for cc in coeffs:
                pcs = np.empty((c, k, d, e), dtype=cc.dtype)
                for i in range(k): pcs[:, i, :, :] = cc[inds, i, :]
                cc = pcs
                poly_coeffs.append(cc)
        else:
            poly_coeffs = np.empty((c, k, d, e), dtype=coeffs.dtype)
            for i in range(k): poly_coeffs[:, i, :, :] = coeffs[inds, i, :]

        # compute extra alpha contrib to integrals
        subalphas = alphas[inds] # c x d array of non-zero alphas

        contribs = []
        for exps, splits, splits_inds, weights in zip(exponents_list, splits_list, splits_inds_list, weights_list):
            poly_chunks = self.evaluate_poly_chunks(
                poly_coeffs, exps, splits, splits_inds, weights, subalphas
            )
            contribs.append(poly_chunks)

        return contribs


    duschinsky_cutoff = 1e-20
    evaluator_plans = {}
    @classmethod
    def evaluate_shifted_poly_overlap(self,
                                      poly:'HermiteProductPolynomial',
                                      Q,
                                      alphas,
                                      zpe_prod,
                                      duschinsky_cutoff=None
                                      ):
        if duschinsky_cutoff is None: duschinsky_cutoff = self.duschinsky_cutoff

        contrib = 0

        work_queues = {}
        max_exponent = max([0] + [
            np.max(e) if len(e) > 0 else 0
            for e, _, _ in poly.term_iter(filter=lambda x,nzi,nzc:sum(x[i] for i in nzi)%2==0)
        ])
        if Q is not None:
            Q_sels = Q[:, :, np.newaxis] ** np.arange(max_exponent+1)[np.newaxis, np.newaxis, :]
        else:
            Q_sels = None
        for exponents, scaling, inds in poly.term_iter(filter=lambda x,nzi,nzc:sum(x[i] for i in nzi)%2==0): # even/odd
            # scaling = scaling
            exponents = np.asanyarray(exponents)
            ordering = np.argsort(-exponents)
            exponents = exponents[ordering]
            inds = [inds[o] for o in ordering]
            poly_coeffs = Q_sels[:, inds, :] if Q is not None else None

            # put coeffs on the queue so we can batch the calls
            key = tuple(exponents)
            queue = work_queues.get(key, None)
            if queue is None:
                queue = []
                work_queues[key] = queue
            queue.append([poly_coeffs, scaling])

        for key,queue in work_queues.items():
            exponents = np.array(key)
            # print("="*10, exponents, "="*10)
            plan = self.evaluator_plans.get(key, None)
            if plan is None:
                if Q is None:
                    plan = self.get_simple_poly_evaluator(exponents)
                else:
                    plan = self.get_poly_evaluation_plan(exponents)
                self.evaluator_plans[key] = plan
            coeff_stack = [poly for poly,scaling in queue if abs(scaling) > duschinsky_cutoff]
            coeff_weights = [scaling for poly,scaling in queue if abs(scaling) > duschinsky_cutoff]
            if len(coeff_weights) > 0:
                subcontribs = plan(alphas, zpe_prod, coeff_stack)
                # print("--->", len(coeff_stack), subcontribs, coeff_weights)
                wtf = sum(w*c for w,c in zip(coeff_weights, subcontribs))
                contrib += wtf
                # print("   >", wtf, contrib)

        return contrib

    @classmethod
    def df_weights(cls, n):
        # gives double factorial weights up to order n
        weights = np.zeros(n+1, dtype=float)
        inds = np.arange(0, n+1, 2)
        double_fac = np.concatenate([[1], np.cumprod(np.arange(1, n, 2))])
        weights[inds] = double_fac
        return weights

    @classmethod
    def get_overlap_gaussian_data(cls,
                                  freqs_gs, modes_gs, inv_gs, center_gs,
                                  freqs_es, modes_es, inv_es, center_es,
                                  rotation_method=None,
                                  rotation_order=None,
                                  rotation_center=None,
                                  include_rotation=None,
                                  rotation_blocks=None
                                  ):

        if include_rotation is None:
            include_rotation = cls.default_include_rotation
        if rotation_order is None:
            rotation_order = cls.default_rotation_order
        if rotation_method is None:
            rotation_method = cls.default_rotation_method
        if rotation_center is None:
            rotation_center = cls.default_rotation_center

        gs_order = State(rotation_order) == State.GroundState
        L_e = np.eye(inv_gs.shape[0])
        L_g = np.eye(inv_gs.shape[0])

        if include_rotation:
            if rotation_blocks is None:
                rotation_blocks = [np.arange(L_e.shape[0])]
            rotation_method = RotationMethods.resolve(rotation_method)
            for block in rotation_blocks:
                if rotation_method == RotationMethods.LeastSquares:
                    # use least squares to express ground state modes as LCs of excited state modes
                    if gs_order:
                        m1 = modes_es[:, block]
                        m2 = modes_gs[:, block]
                    else:
                        m1 = modes_gs[:, block]
                        m2 = modes_es[:, block]
                    ls_tf = np.linalg.inv(m1.T @ m1) @ m1.T @ m2
                    U, s, V = np.linalg.svd(ls_tf)
                    L = U @ V  # Unitary version of a Duschinsky matrix
                    L = L.T
                    if gs_order:
                        l_g = L
                        l_e = np.eye(L.shape[0])
                    else:
                        l_e = L
                        l_g = np.eye(L.shape[0])
                elif rotation_method == RotationMethods.Duschinsky:
                    if gs_order:
                        # del_G E
                        l_g = inv_gs[block, :] @ modes_es[:, block]
                        # l_g = (inv_gs[block, :] @ modes_es[:, block]).T
                        l_e = np.eye(l_g.shape[0])
                    else:
                        # del_E G
                        l_e = inv_es[block, :] @ modes_gs[:, block]
                        # l_e = (inv_gs[block, :] @ modes_es[:, block]).T
                        l_g = np.eye(l_e.shape[0])
                elif callable(rotation_method):
                    l_g, l_e = rotation_method(rotation_order, block)
                else:
                    raise ValueError(f"don't understand rotation method {rotation_method}")
                L_e[np.ix_(block, block)] = l_e
                L_g[np.ix_(block, block)] = l_g

        # find displacement vector (center of gs ground state in this new basis)
        # modes_gs = modes_es @ L

        es_center = State(rotation_center) == State.ExcitedState
        if gs_order:
            tf = modes_es
        else:
            tf = modes_gs
        c_g = center_gs
        c_e = center_es
        if es_center:
            dx_g = (c_g - c_e) @ tf
            dx_e = np.zeros(dx_g.shape)
        else:
            dx_e = (c_e - c_g) @ tf
            dx_g = np.zeros(dx_e.shape)

        # inverse covariance matrices in this coordinate system

        # del_E G or del_G E
        Z_gs = (L_g @ np.diag(freqs_gs) @ L_g.T)
        Z_es = (L_e @ np.diag(freqs_es) @ L_e.T)
        Z_c = Z_gs + Z_es

        # X_g = nput.fractional_power(L_g, -1)
        # print(L_g)
        # print(X_g)
        # X_e = nput.fractional_power(L_e, -1)
        X_g = np.linalg.inv(L_g)
        X_e = np.linalg.inv(L_e)
        S_gs = (X_g.T @ np.diag(1/freqs_gs) @ X_g)
        S_es = (X_e.T @ np.diag(1/freqs_es) @ X_e)
        S_c = S_gs + S_es

        norm_gs = np.power(np.linalg.eigvalsh(Z_gs), 1/4)
        norm_es = np.power(np.linalg.eigvalsh(Z_es), 1/4)
        # prefactor = (
        #         np.prod(norm_gs * norm_es) / np.power(np.pi, len(freqs_es) / 2)
        #         * np.exp(-1/2 * np.dot(dx_g+dx_e, np.dot(np.linalg.inv(S_c), dx_g+dx_e))) # one of these is always zero
        # )
        # prefactor = 1
        decay = (
                np.exp(-1/2 * np.dot(dx_g+dx_e, np.dot(np.linalg.inv(S_c), dx_g+dx_e))) # one of these is always zero
        )
        # print(np.linalg.norm(dx_g+dx_e))
        # print(np.dot(dx_g + dx_e, np.dot(np.linalg.inv(S_c), dx_g + dx_e)))

        alphas_c, modes_c = np.linalg.eigh(Z_c)
        center = np.linalg.inv(Z_c) @ (Z_es @ dx_e + Z_gs @ dx_g)

        zpe_prod = decay * np.prod(norm_gs*norm_es/np.sqrt(alphas_c/2))
        # zpe_prod = prefactor * self.zero_point_alpha_contrib(alphas_c)
        # zpe_prod = decay * self.zero_point_alpha_contrib(norm_gs*norm_es*alphas_c / np.power(np.pi, len(freqs_es) / 2))

        return (alphas_c, modes_c, center, zpe_prod), (L_g, dx_g), (L_e, dx_e)

    integral_block_size = 1000
    @classmethod
    def eval_fcf_overlaps(self,
                          excitations_gs, freqs_gs, modes_gs, inv_gs, center_gs,
                          excitations_es, freqs_es, modes_es, inv_es, center_es,
                          duschinsky_cutoff=None,
                          logger=None,
                          **rotation_opts
                          ):
        """
        Evaluates the Gaussian overlaps between two H.O. wave functions defined by
        a set of polynomial coefficients, broadening factors, and centers, assuming
        the modes and centers are in an Eckart fream
        """

        logger = Logger.lookup(logger)

        # The steps to this process
        # 0. Express ground state nms. w.r.t excited state frame
        # 1. find the overlap Gaussian parameters
        # 2. express the shifts w.r.t the corresponding Gaussian frames
        # 3. shift the polynomials so they are centered at the overlap Gaussian (but don't rotate)
        # 4. express the normal modes (transformation coeffs) w.r.t. the central Gaussian coords
        # 5. for every pair of polynomial terms, if even, evaluate polynomial contrib, reusing when possible

        freqs_gs = np.asanyarray(freqs_gs)
        freqs_es = np.asanyarray(freqs_es)
        modes_gs = np.asanyarray(modes_gs)
        modes_es = np.asanyarray(modes_es)
        inv_gs = np.asanyarray(inv_gs)
        inv_es = np.asanyarray(inv_es)
        center_gs = np.asanyarray(center_gs)
        center_es = np.asanyarray(center_es)

        (alphas, modes_c, center, zpe_prod), (L_gs, c_gs), (L_es, c_es) = self.get_overlap_gaussian_data(
            freqs_gs, modes_gs, inv_gs, center_gs,
            freqs_es, modes_es, inv_es, center_es,
            **rotation_opts
        )

        with logger.block(tag='Duschinksy Transformation'):
            logger.log_print(
                L_gs,
                message_prepper=logger.prep_array
            )
            logger.log_print(
                "disp: {av}",
                av=c_es-c_gs
            )

        # We now have the space to define the parameters that go into the overlap calculation
        shift_gs = center - c_gs #- L_gs @ modes_c.T @ center
        shift_es = center - c_es #- modes_c.T @ center

        Q_gs = (modes_c @ L_gs)
        Q_es = (modes_c @ L_es)

        start = time.time()
        with logger.block(tag="Computing base polynomials"):
            polys1 = [
                HermiteProductPolynomial.from_quanta(eg, freqs_gs).shift(shift_gs)
                for eg in excitations_gs
            ]
            polys2 = [
                HermiteProductPolynomial.from_quanta(ee, freqs_es).shift(shift_es)
                for ee in excitations_es
            ]
            end = time.time()
            logger.log_print("   took: {e:.3f}s", e=(end-start))

        Q = np.concatenate([Q_gs, Q_es], axis=1)


        with logger.block(tag="Evaluating Overlaps"):
            overlaps = []
            n = 0
            block_counter = 1
            block_size = self.integral_block_size
            full_blocks = len(excitations_gs) * len(excitations_es)
            num_blocks = int(np.ceil(full_blocks / block_size))
            block_start = None
            logger.log_print("num integrals: {n}", n=full_blocks)
            for eg,spoly_gs in zip(excitations_gs,polys1):
                subovs = []
                subcounter = 0
                with logger.block(
                        tag="Overlaps from: {gs}",
                        gs=eg,
                        preformatter=lambda **kw:dict(kw, gs=StateMaker.parse_state(kw['gs']))
                    ):
                    for spoly_es in polys2:
                        if n % block_size == 0:
                            if block_start is not None:
                                block_end = time.time()
                                logger.log_print(
                                    "{fmt_table}",
                                    es=excitations_es,
                                    ov=subovs[(subcounter - 1)*block_size:(subcounter)*block_size],
                                    preformatter=lambda **kw: dict(kw,
                                                                   fmt_table=self.format_overlap_tables(
                                                                       kw['es'],
                                                                       kw['ov'],
                                                                       include_headers=False
                                                                   )
                                                                   )
                                )
                                logger.log_print("   took: {e:.3f}s", e=block_end-block_start)
                            block_start = time.time()
                            logger.log_print("evaluating block {} of {}".format(
                                block_counter, num_blocks
                            ))
                            block_counter += 1
                            subcounter += 1
                        n += 1
                        poly = spoly_gs.concat(spoly_es)
                        ov = self.evaluate_shifted_poly_overlap(
                            poly, Q,
                            alphas,
                            zpe_prod,
                            duschinsky_cutoff=duschinsky_cutoff
                        )
                        # print("<::", scaling, ov/scaling, ov)
                        subovs.append(ov)
                logger.log_print(
                    "{fmt_table}",
                    es=excitations_es,
                    ov=subovs,
                    preformatter=lambda **kw:dict(kw,
                                                  fmt_table=self.format_overlap_tables(kw['es'], kw['ov'])
                                                  )
                )

                overlaps.extend(subovs)

        return overlaps

    @classmethod
    def embed_modes(cls, gs_nms: 'NormalModes', es_nms, ref=None, masses=None):
        from ..Molecools import Molecule

        if masses is None:
            masses = gs_nms.masses
        if masses is None:
            masses = es_nms.masses
        if masses is None:
            raise ValueError("need masses to reembed normal modes")
        if ref is None:
            ref = cls.default_embedding_ref

        gs_embedding = State(ref) == State.GroundState

        if not gs_embedding:
            es_nms, gs_nms = gs_nms, es_nms

        mw_gs = gs_nms.mass_weighted
        mw_es = es_nms.mass_weighted
        gs_nms = gs_nms.remove_mass_weighting()
        es_nms = es_nms.remove_mass_weighting()

        gs = Molecule(['H'] * len(masses), gs_nms.origin, masses=masses)

        embedding = gs.get_embedding_data(es_nms.origin)
        ref_coords = embedding.reference_data.coords # embedded in PAF
        emb_ax = 0
        cart_ax = 1
        mat = gs_nms.matrix.reshape(ref_coords.shape + (-1,))
        ref_mat = np.moveaxis(
            np.tensordot(embedding.reference_data.axes, mat, axes=[emb_ax, cart_ax]),
            0, cart_ax
        ).reshape(gs_nms.matrix.shape)
        # emb_ax = (emb_ax + 1) % 2
        # cart_ax = 2
        # mat = gs_nms.inverse.reshape((-1,) + ref_coords.shape)
        # ref_inv = np.moveaxis(
        #     np.tensordot(embedding.reference_data.axes, mat,
        #                  axes=[emb_ax, cart_ax]),
        #     0, cart_ax
        # ).reshape(gs_nms.inverse.shape)
        gs_modes = NormalModes(
            gs_nms.basis,
            ref_mat,
            origin=ref_coords,
            inverse=ref_mat.T / np.repeat(gs_nms.masses, 3)[np.newaxis, :],
            freqs=gs_nms.freqs,
            masses=gs_nms.masses,
            mass_weighted=False
        )

        rot = embedding.rotations[0]
        emb_coords = (embedding.coord_data.coords[0]) @ rot.T # embedded in PAF,
        # raise Exception(emb_coords, ref_coords, rot)
        emb_ax = 0
        cart_ax = 1
        mat = es_nms.matrix.reshape(emb_coords.shape + (-1,))
        rot = embedding.coord_data.axes[0].T
        emb_mat = np.moveaxis(
            np.tensordot(rot, mat, axes=[emb_ax, cart_ax]),
            0, cart_ax
        ).reshape(es_nms.matrix.shape)
        # emb_ax = (emb_ax + 1) % 2
        # cart_ax = 2
        # mat = es_nms.inverse.reshape((-1,) + ref_coords.shape)
        # emb_inv = np.moveaxis(
        #     np.tensordot(rot, mat, axes=[emb_ax, cart_ax]),
        #     0, cart_ax
        # ).reshape(es_nms.inverse.shape)
        es_modes = NormalModes(
            gs_modes.basis,
            emb_mat,
            inverse=emb_mat.T / np.repeat(es_nms.masses, 3)[np.newaxis, :],
            origin=emb_coords,
            freqs=es_nms.freqs,
            masses=es_nms.masses,
            mass_weighted=False
        )

        if mw_gs:
            gs_modes = gs_modes.make_mass_weighted()
        if mw_es:
            es_modes = es_modes.make_mass_weighted()

        if not gs_embedding:
            gs_modes, es_modes = es_modes, gs_modes

        return gs_modes, es_modes

    @classmethod
    def mass_weight_nms(cls, nms, masses=None):
        # if masses is None: masses = nms.masses
        new = nms.make_mass_weighted(masses=masses)
        return new

    @classmethod
    def make_dimensionless(cls, nms, freqs=None, masses=None):
        # if masses is None: masses = nms.masses
        new = nms.make_dimensionless(freqs=freqs, masses=masses)
        return new

    @classmethod
    def prep_states_from_threshold_and_quanta(cls, nms, *,
                                              threshold=None, min_freq=None, max_state=None,
                                              min_quanta=None, max_quanta=None
                                              ):
        from ..BasisReps import BasisStateSpace

        if threshold is None and max_state is None and max_quanta is None:
            raise ValueError("need at least one filter criterion out of `threshold`, `max_state`, or `max_quanta`")

        freqs = nms.freqs
        if threshold is None:
            threshold = freqs[-1] * (np.sum(max_state) if max_state is not None else max_quanta)

        return BasisStateSpace.states_under_freq_threshold(
            freqs,
            threshold,
            min_freq=min_freq,
            max_state=max_state,
            min_quanta=min_quanta,
            max_quanta=max_quanta
        )

    @classmethod
    def prep_states_from_excitations(cls, nms, *, states, **opts):
        from ..BasisReps import StateMaker
        state = StateMaker(len(nms.freqs), **opts)
        return np.array([
            state(*exc)
            for exc in states
        ])

    state_space_prep_registry = {}
    @classmethod
    def state_space_prep_dispatchers(cls):
        return dict(
            {
                'threshold': cls.prep_states_from_threshold_and_quanta,
                'states': cls.prep_states_from_excitations
            },
            **cls.state_space_prep_registry
        )
    @classmethod
    def dispatch_state_space_prep(cls, spec, nms):
        method = spec.get('method', None)
        if method is None:
            # infer method from spec
            if 'threshold' in spec or 'quanta' in spec:
                method = 'threshold'
            elif len(spec.keys() & cls.state_space_prep_dispatchers()) > 0:
                method = next(iter(spec.keys() & cls.state_space_prep_dispatchers()))
            else:
                raise ValueError("can't infer state space method from {}".format(spec))
        spec = spec.copy()
        if 'method' in spec: del spec['method']

        prepper = cls.state_space_prep_dispatchers().get(method)

        return prepper(nms, **spec)

    @classmethod
    def _check_listable(cls, excitations):
        try:
            return all(
                isinstance(exc, np.ndarray) or all(
                    nput.is_numeric(e) or all(nput.is_numeric(ee) for ee in e)
                    for e in exc
                )
                for exc in excitations
            )
        except TypeError:
            return False
    @classmethod
    def prep_state_space(cls, excitations, nms, check=True):
        """
        Dispatcher to get appropriate state spaces
        :param excitations:
        :param check:
        :return:
        """
        if excitations is None: return excitations

        refilter = not check
        if check:
            try:
                exc0 = excitations[0][0]
            except (IndexError, KeyError, TypeError):
                refilter = True
            else:
                refilter = (
                    not isinstance(exc0, int)
                    or any(len(e) < len(nms.freqs) for e in excitations)
                )

        if refilter:
            if isinstance(excitations, dict):
                # dispatch on methods
                excitations = cls.dispatch_state_space_prep(excitations, nms)
            elif isinstance(excitations, (int, np.integer)) or isinstance(excitations[0], (int, np.integer)):
                from ..BasisReps import BasisStateSpace, HarmonicOscillatorProductBasis as HO
                excitations = BasisStateSpace.from_quanta(
                    HO(len(nms.freqs)), excitations
                ).excitations
            elif cls._check_listable(excitations):
                excitations = cls.dispatch_state_space_prep(
                    {
                        "states": excitations
                    }, nms
                )

            else:
                raise ValueError("can't build a state space from {}".format(excitations))

        return excitations


    @classmethod
    def get_fcfs(cls,
                 gs_nms: 'NormalModes', es_nms: 'NormalModes',
                 excitations, ground_states=None,
                 duschinsky_cutoff=None,
                 logger=None,
                 **rotation_embedding_opts
                 ):

        gs_nms, es_nms, rotation_opts = cls.prep_modes(
            gs_nms, es_nms,
            **rotation_embedding_opts
        )
        gs_args, es_args = cls.prep_overlap_args(gs_nms, es_nms)

        ndim = len(gs_nms.freqs)
        if ground_states is None:
            ground_states = [[0] * ndim]

        return cls.eval_fcf_overlaps(
            ground_states, *gs_args,
            excitations, *es_args,
            duschinsky_cutoff=duschinsky_cutoff,
            logger=logger,
            **rotation_opts
        )
    @classmethod
    def format_overlap_tables(cls, es, overlaps, include_headers=True):

        fmt_table = TableFormatter(
            [StateMaker.parse_state, ".3e"],
            headers=["State", "Overlap"] if include_headers else None
        ).format(
            [[s, o] for s, o in zip(es, overlaps)]
        )

        return fmt_table
    @classmethod
    def get_fcf_spectrum(self,
                         gs_nms: 'NormalModes', es_nms: 'NormalModes',
                         excitations, ground_states=None,
                         logger=None,
                         duschinsky_cutoff=None,
                         return_states=False,
                         **rotation_embedding_opts
                         ):
        from ..Spectra import DiscreteSpectrum
        from McUtils.Data import UnitsData

        gs_nms, es_nms, rotation_opts = self.prep_modes(
            gs_nms, es_nms,
            **rotation_embedding_opts
        )


        excitations = self.prep_state_space(excitations, es_nms)
        ground_states = self.prep_state_space(ground_states, gs_nms)

        overlaps = self.get_fcfs(
            gs_nms, es_nms,
            excitations, ground_states=ground_states,
            embed=False, masses=None, mass_weight=False,
            duschinsky_cutoff=duschinsky_cutoff,
            logger=logger,
            **rotation_opts
        )

        freqs = np.dot(np.asanyarray(excitations), es_nms.freqs) * UnitsData.convert("Hartrees", "Wavenumbers")
        ints = np.power(overlaps, 2)

        if ground_states is not None:
            ret = []
            bs = len(freqs)
            for i in range(len(ground_states)):
                ret.append(
                    DiscreteSpectrum(freqs, ints[i*bs:(i+1)*bs])
                )
        else:
            ret = DiscreteSpectrum(freqs, np.power(overlaps, 2))

        if return_states:
            ret = (ret, (ground_states, excitations))

        return ret

    def prep_opts(self, **opts):
        return dict(
            dict(
                dict(
                    logger=self.logger,
                    rotation_order=self.rotation_order if self.rotation_order is not None else self.default_rotation_order,
                    rotation_center=self.rotation_center if self.rotation_center is not None else self.rotation_center,
                    rotation_method=self.rotation_method if self.rotation_method is not None else self.rotation_method,
                    rotation_blocks=self.rotation_blocks,
                    include_rotation=self.include_rotation if self.include_rotation is not None else self.default_include_rotation
                ),
                **opts,
            ),
            embed=False,
            mass_weight=False,
            dimensionless=False,
            mode_reordering=None
        )

    def get_spectrum(self,
                     excitations, *, ground_states=None,
                     return_states=False,
                     duschinsky_cutoff=None,
                     **rotation_embedding_opts
                     ):

        return self.get_fcf_spectrum(
            self.gs_nms, self.es_nms,
            excitations,
            **self.prep_opts(
                return_states=return_states,
                ground_states=ground_states,
                duschinsky_cutoff=duschinsky_cutoff,
                **rotation_embedding_opts
            )
        )

    def get_ezFCF_input(self, excitations, atoms=None, ground_states=None, **rotation_embedding_opts):
        from .ezFCFInterface import ezFCFInterface

        return ezFCFInterface(
            self.atoms if atoms is None else atoms,
            self.gs_nms,
            self.es_nms,
            excitations,
            **{
                k:(v.value if hasattr(v, 'value') else v)
                for k,v in
                self.prep_opts(
                    ground_states=ground_states,
                    # duschinsky_cutoff=duschinsky_cutoff,
                    **rotation_embedding_opts
                ).items()
            }
        )


class HermiteProductPolynomial:

    def __init__(self, poly_dict: dict, ndim):
        self.polys = poly_dict
        self.ndim = ndim

    def shift(self, s):
        return HermiteProductPolynomial(
            {
                k: p.shift(s[k])
                for k, p in self.polys.items()
            },
            self.ndim
        )

    def concat(self, other:'HermiteProductPolynomial'):
        new_poly = self.polys.copy()
        for k,p in other.polys.items():
            new_poly[k + self.ndim] = p
        return HermiteProductPolynomial(
            new_poly,
            self.ndim + other.ndim
        )

    def term_iter(self, filter=None):
        term_coords = list(self.polys.keys())
        term_orders = [
            self.polys[k].coeffs.shape[0]
            for k in term_coords
        ]
        term_inds = [
            k for k, n in enumerate(term_orders)
            if n > 1
        ]
        term_coords = [term_coords[i] for i in term_inds]
        # term_orders = [term_orders[i] for i in term_inds]

        terms_iter = it.product(*[range(o) for o in term_orders])
        for term in terms_iter:
            if filter is None or filter(term, term_inds, term_coords):
                # term_list = sum(([i]*k for i,k in zip(term_coords, term)), [])
                sub_inds = [i for i in term_inds if term[i] > 0]
                term_prod = np.prod([
                    self.polys[i].coeffs[k]
                    for i,k in zip(term_coords, term)
                ])
                term = [term[i] for i in sub_inds]
                yield term, term_prod, [term_coords[i] for i in sub_inds]

    @classmethod
    def from_quanta(cls, quanta, alphas):
        return cls(
            {
                k: cls.get_1D_hermite_poly(n, a)
                for k, (n, a) in enumerate(zip(quanta, alphas))
                if n > 0
            },
            len(alphas)
        )

    _hermite_cache = {}
    @classmethod
    def _hermite_coeffs(cls, n):
        if n not in cls._hermite_cache:
            cls._hermite_cache[n] = np.flip(np.asarray(sp.special.hermite(n, monic=False)))
        return cls._hermite_cache[n]
    @classmethod
    def get_1D_hermite_poly(cls, n, a):
        return DensePolynomial(
            cls._hermite_coeffs(n) * np.sqrt(
                a ** np.arange(n + 1) / (2 ** (n) * math.factorial(n))
            )
        )


"""
Provides a very general specification for a RepresentationBasis object that can be
used to define matrix representations
"""

import abc, numpy as np, scipy.sparse as sp, itertools as ip
from .StateIndexers import BaseStateIndexer, ArrayStateIndexer, PermutationStateIndexer
from McUtils.Scaffolding import MaxSizeCache
from McUtils.Combinatorics import SymmetricGroupGenerator

__all__ = [
    "RepresentationBasis",
    "SimpleProductBasis"
]

__reload_hook__ = ['.StateIndexers', '.Representations', '.Operators']

class RepresentationBasis(metaclass=abc.ABCMeta):
    """
    Metaclass for representations.
    Requires concrete implementations of the position and momentum operators.
    """
    name = "Basis"
    def __init__(self, function_generator, n_quanta):
        """

        :param function_generator:
        :type function_generator:
        :param n_quanta: numbers of quanta (hold over from initial implementation)
        :type n_quanta: int
        """
        self.quanta = n_quanta
        self.generator = function_generator

    @abc.abstractmethod
    def __eq__(self, other):
        raise NotImplementedError("abstract equality not implemented")

    @property
    def dimensions(self):
        """
        Returns the dimensions of the basis

        :return:
        :rtype:
        """
        return (self.quanta,)

    @property
    def ndim(self):
        """
        Returns the number of dimensions of the basis

        :return:
        :rtype:
        """
        return 1

    def ravel_state_inds(self, idx):
        """
        Converts state indices from an array of quanta to an array of indices...except in 1D this really isn't doing anything

        :param idx: indices
        :type idx: Iterable[Iterable[int]]
        :return: array of state indices in the basis
        :rtype: tuple[int]
        """

        idx = np.asarray(idx, dtype=int)
        last_dim = idx.shape[-1]
        if last_dim == 1:
            return idx.reshape(idx.shape[:-1])
        else:
            return idx

    def unravel_state_inds(self, idx):
        """
        Converts state indices from an array of ints to an array of quanta...except in 1D this really isn't doing anything

        :param idx: indices
        :type idx: Iterable[int]
        :return: array of state tuples in the basis
        :rtype: tuple[tuple[int]]
        """

        idx = np.asarray(idx, dtype=int)
        last_dim = idx.shape[-1]
        if last_dim == 1:
            return idx
        else:
            return np.expand_dims(idx, axis=-1)

    def __getitem__(self, item):
        if self.generator is None:
            raise ValueError("basis function generator is None (i.e. explicit functions can't be returned)")
        return self.generator(item)
    def __repr__(self):
        unnamed = self.name == "Basis"
        return "{}({}dimension={})".format(
            type(self).__name__,
            "'{}',".format(self.name) if not unnamed else "",
            self.dimensions#self.generator
        )

    @abc.abstractmethod
    def p(self, n):
        """
        Generates the momentum matrix up to n-quanta.
        There's one big subtlety to what we're doing here, which is that
          for efficiency reasons we return an entirely real matrix
        The reason for that is we assumed people would mostly use it in the context
          of stuff like pp, pQp, or pQQp, in which case the imaginary part pulls out
          and becomes a negative sign
        We actually use this assumption across _all_ of our representations.

        :param n:
        :type n:
        :return:
        :rtype:
        """
        raise NotImplemented

    @abc.abstractmethod
    def x(self, n):
        """
        Generates the coordinate matrix up to n-quanta

        :param n:
        :type n:
        :return:
        :rtype:
        """
        raise NotImplemented

    def I(self, n):
        """
        Generates the identity matrix up to n-quanta

        :param n:
        :type n:
        :return:
        :rtype:
        """
        return sp.csc_matrix(sp.eye(n))

    @property
    def operator_mapping(self):
        return {'x':self.x, 'p':self.p, 'I':self.I}

    selection_rules_mapping = {'x': None, 'p': None, 'I': [0]}
    def operator(self, *terms, logger=None, parallelizer=None, chunk_size=None,
                 **operator_settings
                 ):
        """
        Provides an `Operator` to handle the given terms

        :param terms:
        :type terms:
        :param logger:
        :type logger:
        :param parallelizer:
        :type parallelizer:
        :param chunk_size:
        :type chunk_size:
        :return:
        :rtype:
        """
        from .Operators import Operator
        funcs = [self.operator_mapping[f] if isinstance(f, str) else f for f in terms]
        q = (self.quanta,)
        op = Operator(funcs, q, logger=logger, parallelizer=parallelizer, chunk_size=chunk_size,
                 **operator_settings
                      )
        return op
    def representation(self, *terms, logger=None, name=None, parallelizer=None, chunk_size=None, memory_constrained=False):
        """
        Provides a representation of a product operator specified by `terms`

        :param terms:
        :type terms:
        :return:
        :rtype:
        """
        from .Representations import Representation
        q=self.quanta
        return Representation(self.operator(*terms,
                                            logger=logger, parallelizer=parallelizer, chunk_size=chunk_size),
                              self, name=name, memory_constrained=memory_constrained
                              )

    @classmethod
    def _sel_rules_from_rules(cls, rules, max_len=None):
        """
        We take the combo of the specified rules, where we take successive products of 1D rules with the
        current set of rules following the pattern that
            1. a 1D change can apply to any index in an existing rule
            2. a 1D change can be appended to an existing rule

        We ensure at each step that the rules remain sorted & duplicates are removed so as to keep the rule sets compact.
        This is done in simple python loops, because doing it with arrayops seemed harder & not worth it for a relatively cheap operation.

        :param rules:
        :type rules:
        :return:
        :rtype:
        """
        from collections import OrderedDict

        rules = [np.sort(x) for x in rules]
        ndim = len(rules)
        if max_len is None:
            max_len = ndim

        cur_rules = {(0,) * max_len}
        for r in rules:
            new_rules = set()
            for e in cur_rules:
                for s in r:
                    for i in range(max_len):
                        shift = e[i] + s
                        new = e[:i] + (shift,) + e[i+1:]
                        new = tuple(sorted(new, key=lambda l: -abs(l)*10-(1 if l> 0 else 0)))
                        new_rules.add(new)
                        continue


                        # keep things sorted by absolute value
                        # to cut down on the amount of sorting we have to do
                        # later
                        break_time = e[i] == 0
                        abs_shift = abs(shift)
                        if i > 0 and abs_shift > abs(e[i-1]): # shift this up
                            j = i-1
                            while j > 0 and abs_shift > abs(e[j]):
                                j-=1
                            while j > 0 and abs_shift == abs(e[j]) and shift < 0 and e[j] > 0:
                                j -= 1
                            new = e[:j] + (shift,) + e[j:i] + e[i+1:]
                        elif i < max_len - 1 and abs_shift < abs(e[i+1]):
                            j = i+1
                            while j < max_len-1 and abs_shift < abs(e[j]):
                                j+=1
                            while j < max_len-1 and abs_shift == abs(e[j]) and shift > 0 and e[j] < 0:
                                j += 1
                            new = e[:i] + e[i+1:j] + (shift,) + e[j:]
                        else: # doesn't need to move
                            j = None
                            new = e[:i] + (shift,) + e[i+1:]
                        new_rules.add(new)
                        if break_time:
                            break
            # print(cur_rules)
            cur_rules = new_rules

        cur_rules = np.array(list(cur_rules))
        # print(np.where(cur_rules[:, -1] == -1))
        new_rules = []
        for r in cur_rules:
            w = np.where(r==0)
            if len(w) > 0:
                w = w[0]
            if len(w) == 0:
                new_rules.append(tuple(r))
            else:
                new_rules.append(tuple(r[:w[0]]))

        new_rules = list(sorted(new_rules, key=lambda l:len(l)*100 + sum(l)))
        # print(new_rules)

        return new_rules

    def selection_rule_steps(self, *terms):
        """
        Generates the full set of possible selection rules for terms

        :param terms:
        :type terms:
        :return:
        :rtype:
        """

        # first we get the rules the basis knows about
        rules = [None] * len(terms)
        for i, t in enumerate(terms):
            if t in self.selection_rules_mapping:
                rules[i] = self.selection_rules_mapping[t]
            else:
                raise ValueError("don't know how to infer selection rules from {}".format(
                    t
                ))

        return rules

    def selection_rules(self, *terms):
        """
        Generates the full set of possible selection rules for terms

        :param terms:
        :type terms:
        :return:
        :rtype:
        """

        # first we get the rules the basis knows about
        rules = self.selection_rule_steps(*terms)

        # print(rules)
        return self._sel_rules_from_rules(rules)

class SimpleProductBasis(RepresentationBasis):
    """
    Defines a direct product basis from a 1D basis.
    Mixed product bases aren't currently supported, but this provides
    at least a sample for how that kind of things could be
    generated.
    """

    _indexer_cache = MaxSizeCache()
    array_indexer_cutoff = 0
    def __init__(self, basis_type, n_quanta, indexer=None):
        """
        :param basis_type: the type of basis to do a product over
        :type basis_type: type
        :param n_quanta: the number of quanta for the representations
        :type n_quanta: Iterable[int]
        :param indexer: an object that can turn state specs into indices and indices into state specs
        :type indexer: BaseStateIndexer
        """
        self.basis_type = basis_type
        self.bases = tuple(basis_type(n) for n in n_quanta)
        if indexer is None:
            if self.ndim > self.array_indexer_cutoff: # use the faster version when there are few permutations
                if self.ndim not in self._indexer_cache:
                    indexer = PermutationStateIndexer(self.ndim)
                    self._indexer_cache[self.ndim] = indexer
                else:
                    indexer = self._indexer_cache[self.ndim]
            else:
                dims = self.dimensions
                if dims not in self._indexer_cache:
                    indexer = ArrayStateIndexer(dims)
                    self._indexer_cache[dims] = indexer
                else:
                    indexer = self._indexer_cache[dims]
        self.indexer = indexer
        super().__init__(self.get_function, None)

    def to_state(self, serializer=None):
        return {
            'basis': self.basis_type,
            'quanta': self.quanta,
            'indexer': self.indexer
        }
    @classmethod
    def from_state(cls, data, serializer=None):
        return cls(serializer.deserializer(data['basis']),
                   serializer.deserializer(data['quanta']),
                   indexer=serializer.deserializer(data['indexer']))

    @property
    def ndim(self):
        """
        Provides the number of dimensions of the basis

        :return:
        :rtype:
        """
        return len(self.bases)

    @property
    def dimensions(self):
        """
        Provides the dimensions of the basis

        :return:
        :rtype:
        """
        return self.quanta

    def __eq__(self, other):
        """
        :param other:
        :type other: SimpleProductBasis
        :return:
        :rtype:
        """
        return all(s==o for s,o in zip(self.bases, other.bases))

    @property
    def quanta(self):
        """
        Provides the quanta in each dimension of the basis

        :return:
        :rtype:
        """
        return tuple(b.quanta for b in self.bases)
    @quanta.setter
    def quanta(self, n):
        if n is not None:
            raise ValueError("{}: '{}' can't be set directly".format(
                type(self).__name__,
                'quanta'
            ))

    @property
    def selection_rules_mapping(self):
        return self.bases[0].selection_rules_mapping

    def ravel_state_inds(self, idx):
        """
        Converts state indices from an array of quanta to an array of indices

        :param idx: indices
        :type idx: Iterable[Iterable[int]]
        :return: array of state indices in the basis
        :rtype: tuple[int]
        """

        return self.indexer.to_indices(idx)

    def unravel_state_inds(self, idx):
        """
        Converts state indices from an array of ints to an array of quanta

        :param idx: indices
        :type idx: Iterable[int]
        :return: array of state tuples in the basis
        :rtype: tuple[tuple[int]]
        """

        return self.indexer.from_indices(idx)

        # return np.array(np.unravel_index(idx, self.quanta)).T

    def get_function(self, idx):
        fs = tuple(b[n] for b, n in zip(self.bases, idx))
        return lambda *r, _fs=fs, **kw: np.prod(f(*r, **kw) for f in _fs)

    def operator(self, *terms, coeffs=None, axes=None,
                 parallelizer=None, logger=None, chunk_size=None,
                 **operator_settings
                 ):
        """
        Builds an operator based on supplied terms, remapping names where possible.
        If `coeffs` or `axes` are supplied, a `ContractedOperator` is built.

        :param terms:
        :type terms:
        :param coeffs:
        :type coeffs:
        :param axes:
        :type axes:
        :return:
        :rtype:
        """
        from .Operators import Operator, ContractedOperator

        funcs = [self.bases[0].operator_mapping[f] if isinstance(f, str) else f for f in terms]
        sel_rules = [self.bases[0].selection_rules_mapping[f] if isinstance(f, str) else None for f in terms]
        if any(s is None for s in sel_rules):
            sel_rules = None
        else:
            sel_rules = [np.asarray(s) for s in sel_rules]
        q = self.quanta
        # determine the symmetries up front to make stuff faster
        ids = [hash(f) for f in terms]
        mapping = {k:i for i,k in enumerate(ids)}
        labels = [mapping[k] for k in ids]
        if coeffs is not None:
            op = ContractedOperator(coeffs, funcs, q, axes=axes, symmetries=labels,
                                    selection_rules=sel_rules,
                                    parallelizer=parallelizer,
                                    logger=logger,
                                    chunk_size=chunk_size,
                                    **operator_settings
                                    )
        else:
            op = Operator(funcs, q, symmetries=labels, selection_rules=sel_rules,
                          parallelizer=parallelizer, logger=logger, chunk_size=chunk_size,
                          **operator_settings
                          )
        return op
    def representation(self, *terms,
                       coeffs=None, name=None, axes=None,
                       logger=None, parallelizer=None, chunk_size=None, memory_constrained=False,
                       **operator_settings
                       ):
        """
        Provides a representation of a product operator specified by _terms_.
        If `coeffs` or `axes` are supplied, a `ContractedOperator` is built.

        :param terms:
        :type terms:
        :return:
        :rtype:
        """
        from .Representations import Representation
        return Representation(
            self.operator(*terms, coeffs=coeffs, axes=axes,
                          parallelizer=parallelizer, logger=logger, chunk_size=chunk_size,
                          **operator_settings
                          ),
            self, logger=logger, name=name, memory_constrained=memory_constrained
        )
    def x(self, n):
        """
        Returns the representation of x in the multi-dimensional basis with every term evaluated up to n quanta
        Whether this is what we want or not is still TBD
        :param n:
        :type n:
        :return:
        :rtype:
        """
        return self.representation(self.bases[0].x)[:n, :n]
    def p(self, n):
        """
        Returns the representation of p in the multi-dimensional basis with every term evaluated up to n quanta
        Whether this is what we want or not is still TBD
        :param n:
        :type n:
        :return:
        :rtype:
        """
        return self.representation(self.bases[0].p)[:n, :n]

    def take_subdimensions(self, dims):
        """
        Casts down to lower dimensional space

        :param dims:
        :type dims:
        :return:
        :rtype:
        """
        qq = self.quanta
        return type(self)(self.basis_type, [qq[d] for d in dims])

    def get_state_space(self, quanta):
        from .StateSpaces import BasisStateSpace
        return BasisStateSpace.from_quanta(self, quanta)



"""
BasisReps manages useful functions for generating & working with basis-set representations of data
"""

__all__ = []
from .Bases import *; from .Bases import __all__ as exposed
__all__ += exposed
from .Operators import *; from .Operators import __all__ as exposed
__all__ += exposed
from .Representations import *; from .Representations import __all__ as exposed
__all__ += exposed
from .HarmonicOscillator import *; from .HarmonicOscillator import __all__ as exposed
__all__ += exposed
from .ClassicalBases import *; from .ClassicalBases import __all__ as exposed
__all__ += exposed
from .Wavefunctions import *; from .Wavefunctions import __all__ as exposed
__all__ += exposed
from .StateSpaces import *; from .StateSpaces import __all__ as exposed
__all__ += exposed
from .StateIndexers import *; from .StateIndexers import __all__ as exposed
__all__ += exposed
from .StateFilters import *; from .StateFilters import __all__ as exposed
__all__ += exposed
from .Util import *; from .Util import __all__ as exposed
__all__ += exposed


"""
Provides basic support for vibronic coupling models
"""

__all__ = []
from .FCFs import *; from .FCFs import __all__ as exposed
__all__ += exposed
from .ezFCFInterface import *; from .ezFCFInterface import __all__ as exposed
__all__ += exposed


"""
Provides representations based on starting from a harmonic oscillator basis
"""

__all__ = [
    "HarmonicOscillatorBasis",
    "HarmonicOscillatorProductBasis",
    "HarmonicOscillatorMatrixGenerator",
    "HarmonicOscillatorRaisingLoweringPolyTerms"
]

import scipy.sparse as sp, numpy as np, functools, itertools, math
import scipy.signal

from .Bases import *
from .Operators import Operator, ContractedOperator
from .StateSpaces import BraKetSpace

from McUtils.Data import WavefunctionData
from McUtils.Scaffolding import MaxSizeCache
import McUtils.Numputils as nput
from McUtils.Combinatorics import StirlingS1, Binomial
from McUtils.Zachary import DensePolynomial

class HarmonicOscillatorBasis(RepresentationBasis):
    """
    Provides a concrete implementation of RepresentationBasis using the H.O.
    Need to make it handle units a bit better.
    Currently 1D, need to make multidimensional in the future.
    """
    name='HarmonicOscillator'
    def __init__(self, n_quanta, m=None, re=None, dimensionless=True):
        self.dimensionless = dimensionless
        if not dimensionless and any(x is None for x in (m, re)):
            raise ValueError("if not dimensionless, parameters cannot be 'None'")
        self.data = WavefunctionData['HarmonicOscillator']
        # raise Exception(self.data)
        super().__init__(self.data, n_quanta)

    def __eq__(self, other):
        return (
            isinstance(other, HarmonicOscillatorBasis)
            and (self.dimensionless is other.dimensionless)
            # need to add another check if non-dimensionless basis used
        )

    selection_rules_mapping = {
        'x':[-1, 1],
        'p':[-1, 1],
        "I":[0]
    }

    @functools.lru_cache(maxsize=128)
    def p(self, n):
        mat = self.pmatrix_ho(n)
        if not self.dimensionless:
            raise NotImplementedError
        return mat
    @staticmethod
    def pmatrix_ho(n):
        """
        There's one big subtlety to what we're doing here, which is that
          for efficiency reasons we return an entirely real matrix
        The reason for that is we assumed people would mostly use it in the context
          of stuff like pp, pQp, or pQQp, in which case the imaginary part pulls out
          and becomes a negative sign
        We actually use this assumption across _all_ of our representations
        :param n:
        :type n:
        :return:
        :rtype: sp.csr_matrix
        """
        # the imaginary terms pull out and just become a negative sign
        ar = 1 / np.sqrt(2) * np.sqrt(np.arange(1, n))
        bands = [
            [  ar,  1],
            [ -ar, -1]
        ]
        return sp.csr_matrix(sp.diags([b[0] for b in bands], [b[1] for b in bands]))

    @functools.lru_cache(maxsize=128)
    def x(self, n):
        return self.qmatrix_ho(n)
    @staticmethod
    def qmatrix_ho(n):
        """

        :param n:
        :type n:
        :return:
        :rtype: sp.csr_matrix
        """

        ar = 1 / np.sqrt(2) * np.sqrt(np.arange(1, n))
        bands = [
            [ar,  1],
            [ar, -1]
        ]
        return sp.csr_matrix(sp.diags([b[0] for b in bands], [b[1] for b in bands]))

    def operator(self, *terms, logger=None, parallelizer=None, chunk_size=None,
                 **operator_settings
                 ):
        """
        Builds an operator based on supplied terms, remapping names where possible.
        If `coeffs` or `axes` are supplied, a `ContractedOperator` is built.

        :param terms:
        :type terms:
        :param coeffs:
        :type coeffs:
        :param axes:
        :type axes:
        :return:
        :rtype:
        """

        if all(isinstance(t, str) for t in terms):
            term_eval = HarmonicProductOperatorTermEvaluator(terms).diag
            def computer(idx, term_eval=term_eval): # because of how Operator works...
                if len(idx) > 1:
                    raise ValueError("multidimensional index by one-dimensional basis ?_?")
                return term_eval, term_eval.selection_rules
            return Operator(computer, (self.quanta,), prod_dim=1,
                            logger=logger, parallelizer=parallelizer, chunk_size=chunk_size,
                            **operator_settings
                            )
        else:
            return super().operator(*terms,
                                    logger=logger, parallelizer=parallelizer, chunk_size=chunk_size,
                                    **operator_settings
                                    )

        # funcs = [self.bases[0].operator_mapping[f] if isinstance(f, str) else f for f in terms]
        # q = self.quanta
        # # determine the symmetries up front to make stuff faster
        # ids = [hash(f) for f in terms]
        # mapping = {k:i for i,k in enumerate(ids)}
        # labels = [mapping[k] for k in ids]
        # if coeffs is not None:
        #     op = ContractedOperator(coeffs, funcs, q, axes=axes, symmetries=labels)
        # else:
        #     op = Operator(funcs, q, symmetries=labels)
        # return op

    # def representation(self, *terms):
    #     """
    #     Provides a representation of a product operator specified by 'terms'
    #     :param terms:
    #     :type terms:
    #     :return:
    #     :rtype:
    #     """
    #
    #     q=self.quanta
    #     return TermComputer(self.operator(*terms), q)

    def __repr__(self):
        return "HOBasis({})".format(
            ",".join(str(x) for x in self.dimensions)
        )

class HarmonicOscillatorProductBasis(SimpleProductBasis):

    """
    Tiny, tiny layer on `SimpleProductBasis` that makes use of some analytic work done
    to support representations of `p` and `x`.
    """

    nquant_max = 10 # arbitrary number from the old days of indexing
    def __init__(self, n_quanta, indexer=None):
        if isinstance(n_quanta, int):
            n_quanta = [self.nquant_max] * n_quanta
        super().__init__(HarmonicOscillatorBasis, n_quanta, indexer=indexer)

    def __eq__(self, other):
        return (
            isinstance(other, HarmonicOscillatorProductBasis)
            and (self.ndim == other.ndim)
        )

    def to_state(self, serializer=None):
        return {
            'quanta': self.quanta,
            'indexer': self.indexer
        }
    @classmethod
    def from_state(cls, data, serializer=None):
        return cls(data['quanta'], indexer=serializer.deserialize(data['indexer']))

    def operator(self, *terms, coeffs=None, axes=None, parallelizer=None, logger=None, chunk_size=None, **operator_settings):
        """
        Builds an operator based on supplied terms, remapping names where possible.
        If `coeffs` are supplied, a `ContractedOperator` is built.

        :param terms:
        :type terms:
        :param coeffs:
        :type coeffs:
        :param axes:
        :type axes:
        :return:
        :rtype:
        """

        if all(isinstance(t, str) for t in terms):
            computer = HarmonicProductOperatorTermEvaluator(terms)

            # determine the symmetries up front to make stuff faster
            ids = [hash(f) for f in terms]
            mapping = {k: i for i, k in enumerate(ids)}
            labels = [mapping[k] for k in ids]

            if coeffs is None:
                op = Operator(computer, self.quanta,
                              prod_dim=len(terms),
                              selection_rules=self.selection_rules(*terms),
                              selection_rule_steps=self.selection_rule_steps(*terms),
                              symmetries=labels,
                              logger=logger, parallelizer=parallelizer, chunk_size=chunk_size,
                              **operator_settings
                              )#, axes=axes)
            else:
                op = ContractedOperator(coeffs, computer, self.quanta,
                                        prod_dim=len(terms),
                                        selection_rules=self.selection_rules(*terms),
                                        selection_rule_steps=self.selection_rule_steps(*terms),
                                        symmetries=labels, axes=axes,
                                        logger=logger, parallelizer=parallelizer, chunk_size=chunk_size,
                                        **operator_settings
                                        )
            return op
        else:
            return super().operator(*terms, coeffs=coeffs, axes=axes,
                                    logger=logger, parallelizer=parallelizer, chunk_size=chunk_size)

    def take_subdimensions(self, dims):
        qq = self.quanta
        return type(self)([qq[d] for d in dims])

    def __repr__(self):
        return "HOBasis(dim={})".format(len(self.dimensions))
        #     ",".join(str(x) for x in self.dimensions)
        # )

default_cache_size = 128


class HarmonicOscillatorRaisingLoweringPolyTerms:
    """
    The polynomial induced by _a_ raising operations and _b_ lowering ops
    """
    _stirlings = None
    _binomials = None

    @classmethod
    def get_poly_coeffs(cls, a, b, shift=0):
        if b > a:
            shift = a - b + shift
            a, b = b, a
        prefactor = cls.binom(a + b, a)
        coeffs = prefactor * cls.get_reduced_raising_lowering_coeffs(a, b)
        if shift != 0:
            coeffs = DensePolynomial.compute_shifted_coeffs(coeffs, shift)
        return coeffs

    @classmethod
    def get_direction_change_poly(cls, delta, shift, baseline=0):
        if not nput.is_numeric(delta):
            if nput.is_numeric(baseline): baseline = [baseline] * len(shift)
            return [cls.get_direction_change_poly(d, s, b) for d,s,b in zip(delta, shift, baseline)]

        sqrt_contrib = np.array([1])
        if delta != 0:
            dir_change = np.sign(delta) == -np.sign(shift)  # avoid the zero case
            if dir_change:
                sqrt_contrib = cls.get_sqrt_remainder_coeffs(delta, shift, baseline)
        return sqrt_contrib

    @classmethod
    def get_sqrt_remainder_coeffs(cls, delta, prev, baseline=0):
        # provides rising or falling coeffs with a starting shift
        # by essentially only providing rising coeffs but shifting appropriately
        d = min(abs(delta), abs(prev))
        if delta == 0: return 1
        elif delta < 0:
            # we're falling towards the baseline from above, so we shift our starting point
            # and swap the ordering so we can use a rising fac instead
            prev = prev - d
        # because of the asymmetry of falling and rising
        # we shift our starting point, and then adjust for the baseline shift
        k = prev + 1 + baseline
        if k == 0:
            return np.array([ np.abs(cls.s1(d, j)) for j in range(0, d + 1) ])
        else:
            return np.array([
                sum(
                    cls.binom(j, l) * np.abs(cls.s1(d, j)) * (k ** (j - l))
                    for j in range(l, d+1)
                )
                for l in range(0, d+1)
            ])

    @classmethod
    def s1(cls, i, j):
        if cls._stirlings is None or cls._stirlings.shape[0] <= i or cls._stirlings.shape[0] <= j:
            cls._stirlings = StirlingS1(2 ** int(np.ceil(np.log2(max([64, i, j])))))
        return cls._stirlings[i, j]

    @classmethod
    def binom(cls, i, j):
        if cls._binomials is None or cls._binomials.shape[0] <= i or cls._binomials.shape[0] <= j:
            cls._binomials = Binomial(2 ** int(np.ceil(np.log2(max([64, i, j])))))
        return cls._binomials[i, j]

    @classmethod
    def get_reduced_raising_lowering_coeffs(cls, a, b):
        return np.array([
            sum(
                (cls.s1(b - w, j) * cls.binom(a, w) * cls.binom(b, w) * math.factorial(w) / (2 ** w))
                for w in range(0, b - j + 1)
            )
            for j in range(0, b + 1)
        ])


class HarmonicOscillatorMatrixGenerator:
    """
    1D evaluator for terms looking like `x`, `p`, `q`, etc.
    All of the overall `(-i)^N` info is in the `ProdOp` class that's expected to hold this.
    Only maintains phase info & calculates elements.
    """

    state_cache_size = default_cache_size
    default_evaluator_mode = 'poly'
    def __init__(self, terms, mode=None):
        self.terms = tuple(terms)
        self.N = len(terms)
        self.generators = {}
        if mode is None:
            mode = self.default_evaluator_mode
        self.mode = mode
        if mode == 'poly':
            self.p_pos = None
        else:
            p_pos = []
            for i, o in enumerate(terms):
                if o == "p":
                    p_pos.append(i)
            self.p_pos = tuple(p_pos)
        self._state_group_cache = MaxSizeCache(self.state_cache_size)
        self._poly_cache = {}

    def __repr__(self):
        return "{}({})".format(
            type(self).__name__,
            ", ".join(self.terms)
        )

    _eval_cache_size = default_cache_size
    _evaluators = None
    @classmethod
    def clear_cache(cls):
        cls._evaluators = MaxSizeCache(cls._eval_cache_size)
    @classmethod
    def set_cache_size(cls, new_size):
        cls._eval_cache_size = new_size
        cls._evaluators = cls._evaluators[:new_size]
    @classmethod
    def load_cached(cls, terms):
        if cls._evaluators is None:
            cls.clear_cache() # lol
        if terms in cls._evaluators:
            eval = cls._evaluators[terms]
        else:
            eval = cls(terms)
            cls._evaluators[terms] = eval

        return eval

    @property
    def selection_rules(self):
        return list(range(-self.N, self.N+1, 2))

    def __call__(self, states):
        return self.evaluate_state_terms(states)
        # return self.evaluate_state_terms(states)

    def state_pair_hash(self, states):
        # we use a hash to avoid recomputing harmonic terms
        # whether or not this is the best hash is certainly up for debate
        h1 = states[0]
        h1.flags.writeable = False
        h2 = states[1]
        h2.flags.writeable = False
        # raise Exception(h1.data, h2.data)
        return (hash(h1.data.tobytes()), hash(h2.data.tobytes()))
        # return (len(states[0]), np.max(states[0]), np.min(states[0]), np.max(states[1]), np.min(states[1]),)

    def pull_state_groups(self, states):

        deltas = states[1] - states[0]
        groups, _ = nput.group_by(np.arange(len(states[1])), deltas)
        # delta_vals = np.unique(deltas)
        # delta_sels = [np.where(deltas == a) for a in delta_vals]
        # delta_sels = []
        # reminds = np.arange(len(deltas))
        # for a in delta_vals:
        #     woop = np.where(deltas[reminds] == a)
        #     reminds = np.setdiff1d(reminds, woop[0])
        #     delta_sels.append(woop)
        return groups

    def evaluate_state_terms(self, states, mode=None):
        """
        Evaluates terms coming from different state excitations.
        Doesn't do any pre-filtering, since that's expected to be in the caller.

        :param states:
        :type states:
        :return:
        :rtype:
        """

        if mode is None:
            mode = self.mode

        h = self.state_pair_hash(states)
        # might fuck things up from a memory perspective...
        if h not in self._state_group_cache:
            (delta_vals, delta_sels) = self.pull_state_groups(states)

            biggo = np.zeros(states[0].shape)
            for a, s in zip(delta_vals, delta_sels):
                gen = self.load_generator(a, mode=mode)
                biggo[s] = gen(states[0][s])

            self._state_group_cache[h] = biggo

        return self._state_group_cache[h]

    def load_generator(self, a, mode=None):
        # print(a)
        if a not in self.generators:
            if mode == 'poly':
                gen = self.poly_term_generator(self.terms, a)
            else:
                gen = self.rho_term_generator(a, self.N, self.p_pos)
            self.generators[a] = gen
        else:
            gen = self.generators[a]
        return gen

    @classmethod
    def _enumerate_path(self, steps, cumsteps, delta): # we assume the parity is fine, cumsteps just an optimization
        if len(steps) == 1:
            a = (steps[0]+delta)//2
            b = (steps[0]-delta)//2
            return [[[a, b]]]
        else:
            s = steps[0]
            sT = cumsteps[0]
            min_a = max(0, (delta + (s-sT))//2)
            max_a = min(s, (delta + (s+sT))//2)
            paths = []
            for a in range(min_a, max_a+1):
                b = s - a
                d = delta - (a - b)
                subpaths = [p + [[a, b]] for p in self._enumerate_path(steps[1:], cumsteps[1:], d)]
                paths.extend(subpaths)
            return paths

    @classmethod
    def get_paths(cls, sizes, change):
        cumsums = np.flip(np.cumsum(np.flip(sizes)))
        # print("===", delta, "===")
        return cls._enumerate_path(sizes, cumsums[1:], change)

    @classmethod
    def _build_xp_blocks(self, terms):
        parities = []
        sizes = []
        cur_parity = (-1 if terms[0] == 'p' else 1)
        s = 0
        for t in terms:
            new_parity = (-1 if t == 'p' else 1)
            if new_parity != cur_parity:
                parities.append(cur_parity)
                cur_parity = new_parity
                sizes.append(s)
                s = 1
            else:
                s += 1
        else:
            parities.append(cur_parity)
            sizes.append(s)
        return np.array(sizes), np.array(parities)

    @classmethod
    def get_path_poly(cls, path, parities=None):
        poly_contrib = None
        k = 0
        # print(">>>", path, parities)
        if parities is None:
            parities = [1]*len(path)

        total_parity = 1
        for (a, b), p in zip(reversed(path), parities):
            total_parity *= p ** b
            coeffs = HarmonicOscillatorRaisingLoweringPolyTerms.get_poly_coeffs(a, b, k)
            # print(f" p({a},{b})({k})>", coeffs)
            if poly_contrib is None:
                poly_contrib = coeffs
            else:
                poly_contrib = scipy.signal.convolve(poly_contrib, coeffs)

            d = a - b
            if d != 0:
                dir_change = np.sign(d) == -np.sign(k)  # avoid the zero case
                if dir_change:
                    sqrt_contrib = HarmonicOscillatorRaisingLoweringPolyTerms.get_sqrt_remainder_coeffs(d, k)
                    # print(f" s({a},{b})({k})>", sqrt_contrib)
                    poly_contrib = scipy.signal.convolve(poly_contrib, sqrt_contrib)
            k = k + d
            # print("..>", poly_contrib)
        # print("+->", total_parity)
        poly_contrib *= total_parity
        # print(" > ", poly_contrib)

        return poly_contrib

    _size_blocks_cache = MaxSizeCache()
    @classmethod
    def _get_poly_coeffs(cls, terms, delta):

        if (delta % 2) != len(terms) % 2:
            return 0
        if abs(delta) > len(terms):
            return 0
        if len(terms) == 0:
            return np.array([1]) # just the constant overlap term
        # we know we need to change by delta
        # over
        if terms not in cls._size_blocks_cache:
            cls._size_blocks_cache[terms] = cls._build_xp_blocks(terms)
        sizes, parities = cls._size_blocks_cache[terms]
        # total_parity = (-1)**(np.count_nonzero(parities-1)//2)
        poly = None
        cumsums = np.flip(np.cumsum(np.flip(sizes)))
        # print("===", delta, "===")
        for path in cls._enumerate_path(sizes, cumsums[1:], delta):
            poly_contrib = cls.get_path_poly(path, parities=parities)

            if poly is None:
                poly = poly_contrib
            else:
                if len(poly) < len(poly_contrib):
                    poly = np.pad(poly, [0, len(poly_contrib)-len(poly)])
                elif len(poly_contrib) < len(poly):
                    poly_contrib = np.pad(poly_contrib, [0, len(poly)-len(poly_contrib)])
                poly = poly + poly_contrib

        return poly / np.sqrt(2)**len(terms)

    def poly_coeffs(self, delta, shift=0):
        if (delta, shift) not in self._poly_cache:
            if delta not in self._poly_cache:
                self._poly_cache[delta] = self._get_poly_coeffs(self.terms, delta)
            if shift != 0:
                base_coeffs = self._poly_cache[delta]
                if isinstance(base_coeffs, (int, float, np.integer, np.floating)):
                    self._poly_cache[(delta, shift)] = base_coeffs
                else:
                    self._poly_cache[(delta, shift)] = DensePolynomial.compute_shifted_coeffs(base_coeffs, shift=shift)
            else:
                self._poly_cache[(delta, shift)] = self._poly_cache[delta]
        return self._poly_cache[(delta, shift)]


    @classmethod
    def get_poly_coeffs(cls, terms, delta, shift=0):
        coeffs = cls._get_poly_coeffs(terms, delta)
        if shift != 0:
            coeffs = DensePolynomial.compute_shifted_coeffs(coeffs, shift=shift)
        return coeffs

    @classmethod
    def poly_term_generator(cls, terms, delta, shift=0):
        coeffs = cls.get_poly_coeffs(terms, delta, shift=shift)
        if not isinstance(coeffs, np.ndarray) and coeffs == 0:
            return lambda n: np.zeros(n.shape)
        else:
            orders = np.arange(len(coeffs))[:, np.newaxis]
            dstart = delta+1 if delta < 0 else 1
            dend = 1 if delta < 0 else delta + 1
            if shift != 0:
                return lambda n: np.dot(
                    coeffs,
                    np.power(n[np.newaxis, :], orders)
                ) * np.sqrt(np.prod([n + i + shift for i in range(dstart, dend)], axis=0))
            else:
                return lambda n: np.dot(
                    coeffs,
                    np.power(n[np.newaxis, :], orders)
                ) * np.sqrt(np.prod([n + i for i in range(dstart, dend)], axis=0))

    _partitions_cache = MaxSizeCache()
    @staticmethod
    def _unique_permutations(elements):
        """
        From StackOverflow, an efficient enough
        method to get unique permutations
        :param perms:
        :type perms:
        :return:
        :rtype:
        """

        class unique_element:
            def __init__(self, value, occurrences):
                self.value = value
                self.occurrences = occurrences

        def perm_unique_helper(listunique, result_list, d):
            if d < 0:
                yield tuple(result_list)
            else:
                for i in listunique:
                    if i.occurrences > 0:
                        result_list[d] = i.value
                        i.occurrences -= 1
                        for g in perm_unique_helper(listunique, result_list, d - 1):
                            yield g
                        i.occurrences += 1

        if not hasattr(elements, 'count'):
            elements = list(elements)
        eset = set(elements)
        listunique = [unique_element(i, elements.count(i)) for i in eset]
        u = len(elements)
        return list(sorted(list(perm_unique_helper(listunique, [0] * u, u - 1)), reverse=True))

    @classmethod
    def rho_term_generator(cls, a, N, sel):
        """
        Returns a function to be called on a quantum number to get the coefficient associated with exciting that mode by `a` quanta over
        `N` steps w/ phase info coming from where the momenta-terms are.

        :param a:
        :type a:
        :param N:
        :type N:
        :param i_phase:
        :type i_phase:
        :param is_complex:
        :type is_complex:
        :param sel:
        :type sel:
        :return:
        :rtype:
        """

        if abs(a) > N or (N - a) % 2 != 0:
            def terms(ni):
                if isinstance(ni, (int, np.integer)):
                    return 0
                else:
                    return np.zeros(ni.shape)
            return terms

        if a < 0:
            up_steps = (N - abs(a)) // 2
            down_steps = (N + abs(a)) // 2
        else:
            down_steps = (N - abs(a)) // 2
            up_steps = (N + abs(a)) // 2

        # this can be sped up considerably by using cleverer integer partitioning stuff and whatno
        #TODO: use properer unique permutations code...
        partitions = np.array(cls._unique_permutations(
            [-1] * down_steps
            + [1] * up_steps
        ))
        # np.unique(
        #     np.array(list(
        #         itertools.permutations(
        #             [-1] * down_steps
        #             + [1] * up_steps
        #         )
        #     )),
        #     axis=0
        # )

        # print(a, down_steps, up_steps, partitions)
        # determine the 'phase' of each 'path'
        phases = np.prod(partitions[:, sel], axis=1) / np.sqrt(2) ** N
        # print("   > p", sel, phases)
        # then compute the 'paths' themselves
        unitized = np.abs(np.sign(partitions - 1))
        paths = np.cumsum(partitions, axis=1) + unitized

        def terms(ni, phases=phases, paths=paths):
            return cls.rho(phases, paths, ni)

        return terms

    @classmethod
    def rho(cls, phases, paths, ni):
        """

        :param phases:
        :type phases:
        :param paths:
        :type paths:
        :param ni:
        :type ni:
        :return:
        :rtype:
        """
        # finally define a function that will apply the the "path" to a quantum number and add everything up
        mult = not isinstance(ni, (int, np.integer))
        if mult:
            # print("   ??", ni)
            ni = np.asanyarray(ni)
            # we need to make the broadcasting work
            # requires basically that we copy the shape from ni to the beginning for phases and paths
            # so basically we figure out how many (1) we need to insert at the end of paths
            # and add a (1) to the start of ni
            nidim = ni.ndim
            for i in range(nidim):
                # phases = np.expand_dims(phases, axis=-1)
                paths = np.expand_dims(paths, axis=-1)
            ni = np.expand_dims(ni, 0)

        path_displacements = paths + ni
        path_displacements[path_displacements < 0] = 0
        path_terms = np.sqrt(np.prod(path_displacements, axis=1))

        # return np.sum(phases * path_terms, axis=0)
        if mult:
            return np.dot(phases, path_terms)
        else:
            return phases * path_terms

class HarmonicProductOperatorTermEvaluator:
    """
    A simple class that can be used to directly evaluate any operator built as a product of `p` and `x` terms.
    Automatically dispatches to provide different permutations.
    """

    def __init__(self, terms):
        """
        :param terms: list of 'x' and 'p' terms
        :type terms:
        """
        self.terms = terms

        p_pos = []
        for i, o in enumerate(terms):
            if o not in {"q", "p", "x"}:
                raise ValueError("Don't know what do with operator {}".format(o))
            elif o == "p":
                p_pos.append(i)

        self.p_pos = p_pos

        self.i_phase = (-1) ** (len(p_pos) // 2)  # the -1 terms coming from the complex part of the phase
        # the residual complex part, in general we need this to be zero
        self.is_complex = len(p_pos) % 2 == 1
        if self.is_complex:
            raise ValueError("For simplicity, complex-valued operators not supported right now")

    _operator_cache = None
    _operator_cache_size = default_cache_size
    @classmethod
    def clear_cache(cls):
        HarmonicOscillatorMatrixGenerator.clear_cache()
        cls._operator_cache = MaxSizeCache(cls._operator_cache_size)
    @classmethod
    def set_cache_size(cls, new_size):
        cls._eval_cache_size = new_size
        cls._operator_cache = cls.get_operator_cache()[:new_size]
    @classmethod
    def get_operator_cache(cls):
        if cls._operator_cache is None:
            cls.clear_cache()
        return cls._operator_cache

    def take_suboperator(self, inds):
        """
        Subsamples terms and returns an evaluator for the corresponding operator.
        Basically just determines how `inds` gets broken down into its component pieces.
        These pieces get strung into 1D evaluators and then pushed into a `ProdOp`.


        :param inds:
        :type inds:
        :return:
        :rtype:
        """
        from collections import OrderedDict

        if len(inds) != len(self.terms):
            raise ValueError("Number of indices specified {} and number of terms {} must agree (from {} into {})".format(
                len(inds),
                len(self.terms),
                inds,
                self.terms
            ))

        ind_map = OrderedDict()
        for i, t in zip(inds, self.terms):
            if i in ind_map:
                ind_map[i].append(t)
            else:
                ind_map[i] = [t]

        term_spec = tuple(tuple(t) for t in ind_map.values())

        cache = self.get_operator_cache()
        if term_spec in cache:
            prop = cache[term_spec]
        else:
            terms_1D = [HarmonicOscillatorMatrixGenerator.load_cached(terms) for terms in term_spec]
            uinds = list(ind_map.keys())
            prop = self.ProdOp(terms_1D, uinds, self.i_phase, self.is_complex)
            cache[term_spec] = prop

        return prop
    @property
    def diag(self):
        return self.take_suboperator([0]*len(self.terms))

    def __getitem__(self, inds):
        """
        Syntactic sugar to delegate to `take_suboperator`

        :param inds:
        :type inds:
        :return:
        :rtype:
        """
        return self.take_suboperator(inds)

    def __call__(self, inds):
        """
        Syntactic sugar to play nicely with `Operator`

        :param inds:
        :type inds:
        :return:
        :rtype:
        """
        op = self[inds]
        return op, op.selection_rules

    def __getstate__(self):
        state = self.__dict__.copy()
        # print(state)
        return state

    class ProdOp:
        """
        Simple subclass to manage product operator evaluation
        """
        def __init__(self, ops, uinds, i_phase, is_complex):
            self.ops = ops
            self.uinds = uinds
            self.i_phase = i_phase
            self.is_complex = is_complex
        def __repr__(self):
            return "{}({})".format(
                type(self).__name__,
                "*".join("({})".format(",".join(o.terms)) for o in self.ops)
            )
        def eval_states(self, states):

            if isinstance(states, BraKetSpace): # allows us to use this space as a hashing key
                states = zip(*states.state_pairs)

            chunk = None
            for s, op in zip(states, self.ops):
                # print(op)
                blob = np.asanyarray(op(s))
                if chunk is None:
                    if isinstance(blob, (int, np.integer, float, np.floating)):
                        chunk = np.array([blob])
                    else:
                        chunk = np.asanyarray(blob)
                else:
                    chunk = chunk * blob

            chunk = self.i_phase * chunk
            # if complex...do what?
            return chunk

        def __call__(self, states):
            return self.eval_states(states)

        @property
        def selection_rules(self):
            return [o.selection_rules for o in self.ops]

    def __repr__(self):
        return "{}({})".format("HOTermEvaluator", ", ".join(str(t) for t in self.terms))